from typing import Dict, List, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr

# from xugrid.plot.pyvista import to_pyvista_grid
from xugrid.core.accessorbase import AbstractUgridAccessor
from xugrid.core.wrap import UgridDataArray, UgridDataset
from xugrid.ugrid.ugrid1d import Ugrid1d
from xugrid.ugrid.ugrid2d import Ugrid2d
from xugrid.ugrid.ugridbase import UgridType


class UgridDatasetAccessor(AbstractUgridAccessor):
    def __init__(self, obj: xr.Dataset, grids: Sequence[UgridType]):
        self.obj = obj
        self.grids = grids

    @property
    def grid(self) -> UgridType:
        """
        Returns the single UGRID topology in this dataset. Raises a TypeError if
        the dataset contains more than one topology.
        """
        ngrid = len(self.grids)
        if ngrid == 1:
            return self.grids[0]
        else:
            raise TypeError(
                "Can only access grid topology via `.grid` if dataset contains "
                f"exactly one grid. Dataset contains {ngrid} grids. Use "
                "`.grids` instead."
            )

    @property
    def name(self) -> str:
        """
        Returns name of the single UGRID topology in this dataset. Raises a
        TypeError if the dataset contains more than one topology.
        """
        ngrid = len(self.grids)
        if ngrid == 1:
            return self.grid.name
        else:
            raise TypeError(
                "Can only access grid name via `.name` if dataset contains "
                f"exactly one grid. Dataset contains {ngrid} grids. Use "
                "`.names` instead."
            )

    @property
    def names(self) -> List[str]:
        """Names of all the UGRID topologies in the dataset."""
        return [grid.name for grid in self.grids]

    @property
    def topology(self) -> Dict[str, UgridType]:
        """Mapping from names to UGRID topologies."""
        return {grid.name: grid for grid in self.grids}

    @property
    def bounds(self) -> Dict[str, Tuple]:
        """
        Mapping from grid name to tuple containing ``minx, miny, maxx, maxy``
        values of the grid's node coordinates for every grid in the dataset.
        """
        return {grid.name: grid.bounds for grid in self.grids}

    @property
    def total_bounds(self) -> Tuple:
        """
        Returns a tuple containing ``minx, miny, maxx, maxy`` values for the
        bounds of the dataset as a whole. Currently does not check whether the
        coordinate reference systems (CRS) of the grids in the dataset match.
        """
        bounds = np.column_stack(list(self.bounds.values()))
        return (
            bounds[0].min(),
            bounds[1].min(),
            bounds[2].max(),
            bounds[3].max(),
        )

    def rename(self, new_name_or_name_dict: Union[str, Dict[str, str]]) -> UgridDataset:
        """
        Give a new name to the UGRID topology and update the associated
        coordinate and dimension names in the Dataset.

        Parameters
        ----------
        new_name_or_name_dict: str or dict
            If the argument is a string, the new name of the topology. This
            only works if the dataset contains a single UGRID topology. If the
            argument is a dict, it used as a mapping from old names to new
            names.
        """
        if isinstance(new_name_or_name_dict, str):
            ngrid = len(self.grids)
            if ngrid != 1:
                raise TypeError(
                    "Can only rename with a single name if dataset contains "
                    f"exactly one grid. Dataset contains {ngrid} grids. Provide "
                    "a dictionary of old name to new name instead."
                )

            name = new_name_or_name_dict
            new_grid, names = self.grid.rename(name, return_name_dict=True)
            new_grids = [new_grid]

        elif isinstance(new_name_or_name_dict, dict):
            names = {}
            new_grids = []
            for grid in self.grids:
                name = new_name_or_name_dict.get(grid.name)
                if name is None:
                    new_grid = grid
                else:
                    new_grid, name_dict = grid.rename(name, return_name_dict=True)
                    names.update(name_dict)
                new_grids.append(new_grid)

        else:
            raise TypeError(
                "new_name_or_name_dict should be str or dict, received instead: "
                f"{type(new_name_or_name_dict).__name__}"
            )

        obj = self.obj
        to_rename = tuple(obj.data_vars) + tuple(obj.coords) + tuple(obj.dims)
        new_obj = obj.rename({k: v for k, v in names.items() if k in to_rename})
        return UgridDataset(new_obj, new_grids)

    def assign_node_coords(self) -> UgridDataset:
        """
        Assign node coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Returns
        -------
        assigned: UgridDataset
        """
        result = self.obj
        for grid in self.grids:
            result = grid.assign_node_coords(result)
        return UgridDataset(result, self.grids)

    def assign_edge_coords(self) -> UgridDataset:
        """
        Assign edge coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Returns
        -------
        assigned: UgridDataset
        """
        result = self.obj
        for grid in self.grids:
            result = grid.assign_edge_coords(result)
        return UgridDataset(result, self.grids)

    def assign_face_coords(self) -> UgridDataset:
        """
        Assign face coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Returns
        -------
        assigned: UgridDataset
        """
        result = self.obj
        for grid in self.grids:
            if grid.topology_dimension > 1:
                result = grid.assign_face_coords(result)
        return UgridDataset(result, self.grids)

    def set_node_coords(self, node_x: str, node_y: str, topology: str = None):
        """
        Given names of x and y coordinates of the nodes of an object, set them
        as the coordinates in the grid.

        Parameters
        ----------
        node_x: str
            Name of the x coordinate of the nodes in the object.
        node_y: str
            Name of the y coordinate of the nodes in the object.
        topology: str, optional
            Name of the grid topology in which to set the node_x and node_y
            coordinates. Can be omitted if the UgridDataset contains only a
            single grid.
        """
        if topology is None:
            if len(self.grids) == 1:
                grid = self.grids[0]
            else:
                raise ValueError(
                    "topology must be specified when dataset contains multiple datasets"
                )
        else:
            grids = [grid for grid in self.grid if grid.name == topology]
            if len(grids != 1):
                raise ValueError(
                    f"Expected one grid with topology {topology}, found: {len(grids)}"
                )
            grid = grids[0]

        grid.set_node_coords(node_x, node_y, self.obj)

    def sel(self, x=None, y=None):
        new_obj = self.obj
        new_grids = []
        for grid in self.grids:
            result = grid.sel(new_obj, x, y)
            if isinstance(result, tuple):
                new_obj, new_grid = result
                new_grids.append(new_grid)
            else:
                new_obj = result

        if new_grids:
            return UgridDataset(new_obj, new_grids)
        else:
            return result

    def sel_points(
        self, x, y, method=None, out_of_bounds="warn", fill_value=np.nan, tolerance=None
    ):
        """
        Select points in the unstructured grid.

        For data on the core dimension (faces for Ugrid2d, edges for Ugrid1d),
        points are located by checking containment unless method is "nearest".
        For data on other dimension (e.g. nodes), the nearest entity is found.

        Parameters
        ----------
        x: 1d array of floats with shape ``(n_points,)``
        y: 1d array of floats with shape ``(n_points,)``
        obj: xr.DataArray or xr.Dataset
        method: str, {None, "nearest"}, optional
            * None: default, locate the core entity (face for Ugrid2d, edge for
              Ugrid1d) that contains each point. For secondary entities (nodes
              and edges for Ugrid2d; nodes for Ugrid1d), the nearest is always returned.
            * "nearest": locate the nearest entity to each point, including the core
              entity.

        out_of_bounds: str, default ``"warn"``
            What to do when points are located outside of any feature:

            * raise: raise a ValueError.
            * ignore: return ``fill_value`` for the out of bounds points.
            * warn: give a warning and return NaN for the out of bounds points.
            * drop: drop the out of bounds points. They may be identified
              via the ``index`` coordinate of the returned selection.
        fill_value: scalar, DataArray, Dataset, or callable, optional, default: np.nan
            Value to assign to out-of-bounds points if out_of_bounds is warn
            or ignore. Forwarded to xarray's ``.where()`` method.
        tolerance: float, optional
            The tolerance used to determine whether a point is on an edge. This
            is a floating point precision criterion, thus cannot be directly be
            interpreted as a distance. If None, ``numba_celltree`` estimates an
            appropriate tolerance by multiplying the maximum diagonal of the
            bounding boxes with 1e-12.

        Returns
        -------
        selection: xr.DataArray or xr.Dataset
            The name of the topology is prefixed in the x, y coordinates
            and in a points dimension.
        """
        result = self.obj
        for grid in self.grids:
            result = grid.sel_points(
                result, x, y, method, out_of_bounds, fill_value, tolerance
            )
        return result

    def rasterize(self, resolution: float) -> xr.Dataset:
        """
        Rasterize all face data on 2D unstructured grids by sampling.

        Parameters
        ----------
        resolution: float
            Spacing in x and y.

        Returns
        -------
        rasterized: xr.Dataset
        """
        datasets = []
        for grid in self.grids:
            xx, yy, index = grid.rasterize(resolution, self.total_bounds)
            datasets.append(self._raster(xx, yy, index))
        return xr.merge(datasets)

    def rasterize_like(self, other: Union[xr.DataArray, xr.Dataset]) -> xr.Dataset:
        """
        Rasterize unstructured all face data on 2D unstructured grids by
        sampling on the x and y coordinates of ``other``.

        Parameters
        ----------
        resolution: float
            Spacing in x and y.
        other: Union[xr.DataArray, xr.Dataset]
            Object to take x and y coordinates from.

        Returns
        -------
        rasterized: xr.Dataset
        """
        x = other["x"].to_numpy()
        y = other["y"].to_numpy()
        datasets = []
        for grid in self.grids:
            xx, yy, index = grid.rasterize_like(x, y)
            datasets.append(self._raster(xx, yy, index))
        return xr.merge(datasets)

    def to_periodic(self):
        """
        Convert every grid to a periodic grid, where the rightmost boundary
        shares its nodes with the leftmost boundary.

        Returns
        -------
        periodic: UgridDataset
        """
        grids = []
        result = self.obj
        for grid in self.grids:
            new_grid, result = grid.to_periodic(obj=result)
            grids.append(new_grid)
        return UgridDataset(result, grids)

    def to_nonperiodic(self, xmax: float):
        """
        Convert the grid from a periodic grid (where the rightmost boundary shares its
        nodes with the leftmost boundary) to an aperiodic grid, where the leftmost nodes
        are separate from the rightmost nodes.

        Parameters
        ----------
        xmax: float
            The x-value of the newly created rightmost boundary nodes.

        Returns
        -------
        nonperiodic: UgridDataset
        """
        grids = []
        result = self.obj
        for grid in self.grids:
            new_grid, result = grid.to_nonperiodic(xmax=xmax, obj=result)
            grids.append(new_grid)
        return UgridDataset(result, grids)

    def intersect_line(
        self, start: Sequence[float], end: Sequence[float]
    ) -> xr.Dataset:
        """
        Intersect a line with the grid of this data, and fetch the values of
        the intersected faces.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset
        start: sequence of two floats
            coordinate pair (x, y), designating the start point of the line.
        end: sequence of two floats
            coordinate pair (x, y), designating the end point of the line.

        Returns
        -------
        intersection: xr.Dataset
            The name of the topology is prefixed in the x, y and s
            (spatium=distance) coordinates.
        """
        obj = self.obj
        for grid in self.grids:
            obj = grid.intersect_line(obj, start, end)
        return obj

    def intersect_linestring(self, linestring) -> xr.Dataset:
        """
        Intersect the grid along a collection of linestrings. Returns a new Dataset
        with the values for each intersected segment.

        Parameters
        ----------
        linestring: shapely.LineString

        Returns
        -------
        intersection: xr.Dataset
            The name of the topology is prefixed in the x, y and s
            (spatium=distance) coordinates.
        """
        obj = self.obj
        for grid in self.grids:
            obj = grid.intersect_linestring(obj, linestring)
        return obj

    def to_dataset(self, optional_attributes: bool = False):
        """
        Convert this UgridDataset into a standard
        xarray.Dataset.

        The UGRID topology information is added as standard data variables.

        Parameters
        ----------
        optional_attributes: bool, default: False.
            Whether to generate the UGRID optional attributes.

        Returns
        -------
        dataset: UgridDataset
        """
        return xr.merge(
            [grid.to_dataset(self.obj, optional_attributes) for grid in self.grids]
        )

    @property
    def crs(self):
        """
        The Coordinate Reference System (CRS) represented as a ``pyproj.CRS`` object.

        Returns None if the CRS is not set.

        Returns
        -------
        crs: dict
            A dictionary containing the names of the grids and their CRS.
        """
        return {grid.name: grid.crs for grid in self.grids}

    def set_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
        allow_override: bool = False,
        topology: str = None,
    ):
        """
        Set the Coordinate Reference System (CRS) of a UGRID topology.

        NOTE: The underlying geometries are not transformed to this CRS. To
        transform the geometries to a new CRS, use the ``to_crs`` method.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted
            by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying the projection.
        allow_override : bool, default False
            If the the UGRID topology already has a CRS, allow to replace the
            existing CRS, even when both are not equal.
        topology: str, optional
            Name of the grid topology in which to set the CRS.
            Sets the CRS for all grids if left unspecified.
        """
        if topology is None:
            grids = self.grids
        else:
            names = [grid.name for grid in self.grids]
            if topology not in names:
                raise ValueError(f"{topology} not found. Expected one of: {names}")
            grids = [grid for grid in self.grids if grid.name == topology]

        for grid in grids:
            grid.set_crs(crs, epsg, allow_override)
            grid._update_coordinate_attrs(self.obj)
        return

    def to_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
        topology: str = None,
    ) -> UgridDataset:
        """
        Transform geometries to a new coordinate reference system.
        Transform all geometries in an active geometry column to a different coordinate
        reference system. The ``crs`` attribute on the current Ugrid must
        be set. Either ``crs`` or ``epsg`` may be specified for output.

        This method will transform all points in all objects. It has no notion
        of projecting the cells. All segments joining points are assumed to be
        lines in the current projection, not geodesics. Objects crossing the
        dateline (or other projection boundary) will have undesirable behavior.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted by
            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying output projection.
        topology: str, optional
            Name of the grid topology to reproject.
            Reprojects all grids if left unspecified.

        Notes
        -----
        Node coordinates are always recomputed from the reprojected node positions.
        Face and edge coordinates, if present, are also recomputed from the new
        node positions. Coordinates not governed by the UGRID conventions
        are left untouched.
        """
        obj = self.obj.copy()

        names = [grid.name for grid in self.grids]
        if topology is not None and topology not in names:
            raise ValueError(f"{topology} not found. Expected one of: {names}")

        grids = []
        for grid in self.grids:
            if topology is None or grid.name == topology:
                newgrid = grid.to_crs(crs, epsg)
                obj = newgrid._assign_derived_coords(obj)
            else:
                newgrid = grid.copy()
            grids.append(newgrid)

        return UgridDataset(obj, grids)

    def to_geodataframe(self, dim_order=None) -> "geopandas.GeoDataFrame":  # type: ignore # noqa
        """
        Convert data and topology of one facet (node, edge, face) of the grid
        to a geopandas GeoDataFrame. This also determines the geometry type of
        the geodataframe:

        * node: point
        * edge: line
        * face: polygon

        Parameters
        ----------
        name: str
            Name to give to the array (required if unnamed).
        dim_order:
            Hierarchical dimension order for the resulting dataframe. Array content is
            transposed to this order and then written out as flat vectors in contiguous
            order, so the last dimension in this list will be contiguous in the resulting
            DataFrame. This has a major influence on which operations are efficient on the
            resulting dataframe.

            If provided, must include all dimensions of this DataArray. By default,
            dimensions are sorted according to the DataArray dimensions order.

        Returns
        -------
        geodataframe: gpd.GeoDataFrame
        """
        import geopandas as gpd

        ds = self.obj

        gdfs = []
        for grid in self.grids:
            if grid.topology_dimension == 1:
                dim = grid.edge_dimension
            elif grid.topology_dimension == 2:
                dim = grid.face_dimension
            else:
                raise ValueError("invalid topology dimension on grid")

            variables = [var for var in ds.data_vars if dim in ds.variables[var].dims]
            if variables:
                data = ds[variables].to_dataframe(dim_order=dim_order)
            else:
                data = None

            # TODO deal with time-dependent data, etc.
            # Basically requires checking which variables are static, which aren't.
            # For non-static, requires repeating all geometries.
            # Call reset_index on multi-index to generate them as regular columns.
            gdfs.append(
                gpd.GeoDataFrame(
                    data=data,
                    geometry=grid.to_shapely(dim),
                    crs=grid.crs,
                )
            )

        return pd.concat(gdfs)

    def reindex_like(
        self,
        other: Union[UgridType, UgridDataArray, UgridDataset],
        tolerance: float = 0.0,
    ):
        """
        Conform this object to match the topology of another object. The
        topologies must be exactly equivalent: only the order of the nodes,
        edges, and faces may differ.

        Topologies are matched by name, and dimension names must match for
        equivalent topologies.

        Parameters
        ----------
        other: Ugrid1d, Ugrid2d, UgridDataArray, UgridDataset
        obj: DataArray or Dataset
        tolerance: float, default value 0.0.
            Maximum distance between inexact coordinate matches.

        Returns
        -------
        reindexed: UgridDataset
        """
        if isinstance(other, (Ugrid1d, Ugrid2d)):
            other_grids = [other]
        elif isinstance(other, (UgridDataArray, UgridDataset)):
            other_grids = other.ugrid.grids
        else:
            raise TypeError(
                "Expected Ugrid1d, Ugrid2d, UgridDataArray, or UgridDataset,"
                f"received instead: {type(other).__name__}"
            )
        # Convert to dict to match by name
        other_grids = {grid.name: grid for grid in other_grids}

        new_grids = []
        result = self.obj
        for grid in self.grids:
            other = other_grids.get(grid.name)
            if other:
                result = grid.reindex_like(other, obj=result, tolerance=tolerance)
                new_grids.append(other)
            else:
                new_grids.append(grid)
        return UgridDataset(result, new_grids)



"""
This module deals with extracting the relevant data from the UGRID attributes.

It takes some inspiration from: https://github.com/xarray-contrib/cf-xarray
"""

import warnings
from collections import ChainMap
from itertools import chain
from typing import Dict, List, Optional, Set, Tuple

import xarray as xr


class UgridDimensionError(Exception):
    pass


class UgridCoordinateError(Exception):
    pass


_DIM_NAMES = {
    1: ("node_dimension", "edge_dimension"),
    2: ("node_dimension", "face_dimension", "edge_dimension"),
}

_COORD_NAMES = {
    1: ("node_coordinates", "edge_coordinates"),
    2: ("node_coordinates", "face_coordinates", "edge_coordinates"),
}
_COORD_DIMS = {
    "node_coordinates": "node_dimension",
    "edge_coordinates": "edge_dimension",
    "face_coordinates": "face_dimension",
}

_CONNECTIVITY_NAMES = {
    1: ("edge_node_connectivity",),
    2: (
        "face_node_connectivity",
        "edge_node_connectivity",
        "face_edge_connectivity",
        "face_face_connectivity",
        "edge_face_connectivity",
        "boundary_node_connectivity",
    ),
}

_CONNECTIVITY_DIMS = {
    "face_node_connectivity": ("face_dimension", None),
    "edge_node_connectivity": ("edge_dimension", 2),
    "face_edge_connectivity": ("face_dimension", None),
    "face_face_connectivity": ("face_dimension", None),
    "edge_face_connectivity": ("edge_dimension", 2),
    "boundary_node_connectivity": ("boundary_edge_dimension", 2),
}

X_STANDARD_NAMES = ("projection_x_coordinate", "longitude")
Y_STANDARD_NAMES = ("projection_y_coordinate", "latitude")

PROJECTED = True
GEOGRAPHIC = False
DEFAULT_ATTRS = {
    "node_x": {
        PROJECTED: {
            "standard_name": "projection_x_coordinate",
        },
        GEOGRAPHIC: {
            "standard_name": "longitude",
        },
    },
    "node_y": {
        PROJECTED: {
            "standard_name": "projection_y_coordinate",
        },
        GEOGRAPHIC: {
            "standard_name": "latitude",
        },
    },
    "edge_x": {
        PROJECTED: {
            "standard_name": "projection_x_coordinate",
        },
        GEOGRAPHIC: {
            "standard_name": "longitude",
        },
    },
    "edge_y": {
        PROJECTED: {
            "standard_name": "projection_y_coordinate",
        },
        GEOGRAPHIC: {
            "standard_name": "latitude",
        },
    },
    "face_x": {
        PROJECTED: {
            "standard_name": "projection_x_coordinate",
        },
        GEOGRAPHIC: {
            "standard_name": "longitude",
        },
    },
    "face_y": {
        PROJECTED: {
            "standard_name": "projection_y_coordinate",
        },
        GEOGRAPHIC: {
            "standard_name": "latitude",
        },
    },
    "face_node_connectivity": {
        "cf_role": "face_node_connectivity",
        "start_index": 0,
        "_FillValue": -1,
    },
    "edge_node_connectivity": {
        "cf_role": "edge_node_connectivity",
        "start_index": 0,
        "_FillValue": -1,
    },
    "face_edge_connectivity": {
        "cf_role": "face_edge_connectivity",
        "start_index": 0,
        "_FillValue": -1,
    },
    "face_face_connectivity": {
        "cf_role": "face_face_connectivity",
        "start_index": 0,
        "_FillValue": -1,
    },
    "edge_face_connectivity": {
        "cf_role": "edge_face_connectivity",
        "start_index": 0,
        "_FillValue": -1,
    },
    "boundary_node_connectivity": {
        "cf_role": "boundary_node_connectivity",
        "start_index": 0,
        "_FillValue": -1,
    },
}


def default_topology_attrs(name: str, topology_dimension: int):
    if topology_dimension == 1:
        return {
            "cf_role": "mesh_topology",
            "long_name": "Topology data of 1D network",
            "topology_dimension": 1,
            "node_dimension": f"{name}_nNodes",
            "edge_dimension": f"{name}_nEdges",
            "edge_node_connectivity": f"{name}_edge_nodes",
            "node_coordinates": f"{name}_node_x {name}_node_y",
            "edge_coordinates": f"{name}_edge_x {name}_edge_y",
        }
    elif topology_dimension == 2:
        return {
            "cf_role": "mesh_topology",
            "long_name": "Topology data of 2D mesh",
            "topology_dimension": 2,
            "node_dimension": f"{name}_nNodes",
            "edge_dimension": f"{name}_nEdges",
            "face_dimension": f"{name}_nFaces",
            "max_face_nodes_dimension": f"{name}_nMax_face_nodes",
            "boundary_edge_dimension": f"{name}_nBoundary_edges",
            "edge_node_connectivity": f"{name}_edge_nodes",
            "face_node_connectivity": f"{name}_face_nodes",
            "face_edge_connectivity": f"{name}_face_edges",
            "edge_face_connectivity": f"{name}_edge_faces",
            "boundary_node_connectivity": f"{name}_boundary_nodes",
            "face_face_connectivity": f"{name}_face_faces",
            "node_coordinates": f"{name}_node_x {name}_node_y",
            "edge_coordinates": f"{name}_edge_x {name}_edge_y",
            "face_coordinates": f"{name}_face_x {name}_face_y",
        }
    else:
        raise ValueError(
            f"topology_dimensions should be 1 or 2, received {topology_dimension}"
        )


def _get_topology(ds: xr.Dataset) -> List[str]:
    return [
        var
        for var in ds.data_vars
        if ds.variables[var].attrs.get("cf_role") == "mesh_topology"
    ]


def _infer_xy_coords(
    ds: xr.Dataset, candidates: List[str]
) -> Tuple[List[str], List[str]]:
    # TODO: add argument for latitude / longitude?
    x = []
    y = []
    for candidate in candidates:
        stdname = ds[candidate].attrs.get("standard_name")
        if stdname in X_STANDARD_NAMES:
            x.append(candidate)
        elif stdname in Y_STANDARD_NAMES:
            y.append(candidate)

    if not x and not y:
        first = candidates[0]
        second = candidates[1]
        warnings.warn(
            f"No standard_name of {X_STANDARD_NAMES + Y_STANDARD_NAMES} in {candidates}.\n"
            f"Using {first} and {second} as projected x and y coordinates.",
            UserWarning,
            stacklevel=2,
        )
        x.append(first)
        y.append(second)
    elif not x:
        raise UgridCoordinateError(
            "No standard_name of {X_STANDARD_NAMES} in {candidates}"
        )
    elif not y:
        raise UgridCoordinateError(
            "No standard_name of {Y_STANDARD_NAMES} in {candidates}"
        )

    return x, y


def _get_coordinates(
    ds: xr.Dataset, topologies: List[str]
) -> Dict[str, Dict[str, Tuple[List[str], List[str]]]]:
    topology_dict = {}
    for topology in topologies:
        attrs = ds[topology].attrs
        topodim = attrs["topology_dimension"]
        vardict = {}
        for name in _COORD_NAMES[topodim]:
            if name in attrs:
                candidates = [c for c in attrs[name].split(" ") if c in ds]
                if len(candidates) == 0:
                    warnings.warn(
                        f"the following variables are specified for UGRID {name}: "
                        f'"{attrs[name]}", but they are not present in the dataset',
                        UserWarning,
                        stacklevel=2,
                    )
                    continue
                if len(candidates) < 2:
                    raise UgridCoordinateError(
                        f"{topology}: at least two values required for UGRID {name},"
                        f' while only "{attrs[name]}" are specified.'
                    )
                vardict[name] = _infer_xy_coords(ds, candidates)

        topology_dict[topology] = vardict

    return topology_dict


def _infer_dims(
    ds: xr.Dataset,
    connectivities: Dict[str, str],
    coordinates: Dict[str, Dict[str, Tuple[List[str]]]],
    vardict: Dict[str, str],
) -> Dict[str, str]:
    """Infer dimensions based on connectivity and coordinates."""
    inferred = {}
    for role, varname in connectivities.items():
        expected_dims = _CONNECTIVITY_DIMS[role]
        var_dims = ds[varname].dims
        for key, dim in zip(expected_dims, var_dims):
            if isinstance(key, str):
                prev_dim = inferred.get(key)
                # Not specified: default order can be used to infer dimensions.
                if prev_dim is None:
                    inferred[key] = dim
                else:
                    if prev_dim not in var_dims:
                        raise UgridDimensionError(
                            f"{key}: {prev_dim} not in {role}: {varname}"
                            f" with dimensions: {var_dims}"
                        )
            elif isinstance(key, int):
                dim_size = ds.sizes[dim]
                if not dim_size == key:
                    raise UgridDimensionError(
                        f"Expected size {key} for dimension {dim} in variable "
                        f"{varname} with role {role}, found instead: "
                        f"{dim_size}"
                    )

            # If key is None, we don't do any checking.

    for role, varnames in coordinates.items():
        key = _COORD_DIMS[role]
        for varname in chain.from_iterable(varnames):
            var_dims = ds[varname].dims
            if len(var_dims) != 1:
                continue
            var_dim = var_dims[0]

            prev_dim = vardict.get(key) or inferred.get(key)
            if prev_dim is None:
                inferred[key] = var_dim
            else:
                if prev_dim != var_dim:
                    raise UgridDimensionError(
                        f"Conflicting names for {key}: {prev_dim} versus {var_dim}"
                    )

    return inferred


def _get_dimensions(
    ds: xr.Dataset,
    topologies: List[str],
    connectivity: Dict[str, Dict[str, str]],
    coordinates: Dict[str, Dict[str, Tuple[List[str]]]],
) -> Dict[str, Dict[str, str]]:
    """
    Get the dimensions from the topology attributes and infer them from
    connectivity arrays or coordinates.
    """
    topology_dict = {}
    for topology in topologies:
        attrs = ds[topology].attrs
        topodim = attrs["topology_dimension"]
        # dimensions are optionally required: only if the dimension order is
        # nonstandard in any of the connectivity variables.
        vardict = {k: attrs[k] for k in _DIM_NAMES[topodim] if k in attrs}
        inferred = _infer_dims(
            ds, connectivity[topology], coordinates[topology], vardict
        )
        topology_dict[topology] = {**vardict, **inferred}

    return topology_dict


def _get_connectivity(
    ds: xr.Dataset, topologies: List[str]
) -> Dict[str, Dict[str, str]]:
    topology_dict = {}
    for topology in topologies:
        attrs = ds[topology].attrs
        topodim = attrs["topology_dimension"]
        topology_dict[topology] = {
            k: attrs[k]
            for k in _CONNECTIVITY_NAMES[topodim]
            if (k in attrs) and (attrs[k] in ds)
        }
    return topology_dict


def _get_grid_mapping_names(
    ds: xr.Dataset,
    topologies: List[str],
    dimensions: Dict[str, Dict[str, str]],
) -> Dict[str, str | None]:
    topology_dict = {}
    varnames = set(ds.variables.keys())
    for topology in topologies:
        topology_dict[topology] = None
        # The grid mapping should be specified per variable.
        # Check which variables have the relevant UGRID dimensions, and extract
        # the grid mapping.
        topo_dims = set(dimensions[topology].values())
        names = {
            var.attrs.get("grid_mapping") or var.encoding.get("grid_mapping")
            for var in ds.variables.values()
            if topo_dims & set(var.dims)
        } - {None}

        if names:
            # In principle, multiple coordinates are allowed to be specified
            # in the topology variable. Let's say there are two sets of coordinates
            # node_coordinates: "mesh2d_node_x1 mesh2d_node_y1 mesh2d_node_x2 mesh2d_node_y2"
            # In this case, we need a grid mapping for (x1, y1) and for (x2, y2),
            # but given that the grid mapping is defined on a data variable, there is no
            # way to link them correctly. Hence the ValueError.
            # See also: https://github.com/ugrid-conventions/ugrid-conventions/issues/64
            if len(names) > 1:
                raise ValueError(
                    f"Multiple grid mappings found for topology '{topology}': "
                    f"{names}. Variables on the same topology are expected to "
                    f"share a single coordinate reference system (CRS). "
                    f"Load the dataset with xarray.open_dataset() and modify "
                    f"the grid_mapping attributes before converting to a "
                    f"UgridDataset."
                )
            name = next(iter(names))
            if name in varnames:
                topology_dict[topology] = name
            else:
                warnings.warn(
                    "The following grid mapping variable is specified in the attribute\n"
                    f"or encoding of one or more variables, but is not present in the dataset: {name}",
                    UserWarning,
                    stacklevel=2,
                )

    return topology_dict


def _infer_projected(
    ds: xr.Dataset,
    topologies: List[str],
    coordinates: Dict[str, Dict[str, Tuple[List[str], List[str]]]],
) -> Dict[str, bool | None]:
    topology_dict = {}
    for topology in topologies:
        inferred = []
        for role, (x_vars, y_vars) in coordinates[topology].items():
            for x_varname, y_varname in zip(x_vars, y_vars):
                # Check x
                stdname = ds[x_varname].attrs.get("standard_name")
                if stdname == X_STANDARD_NAMES[0]:
                    inferred.append((x_varname, True))
                elif stdname == X_STANDARD_NAMES[1]:
                    inferred.append((x_varname, False))
                # Check y
                stdname = ds[y_varname].attrs.get("standard_name")
                if stdname == Y_STANDARD_NAMES[0]:
                    inferred.append((y_varname, True))
                elif stdname == Y_STANDARD_NAMES[1]:
                    inferred.append((y_varname, False))

        # In principle, a geocentric CRS like EPSG:4328 is neither projected
        # nor geographic, but it is very niche we cannot easily support it
        # within xugrid.
        values = {v for _, v in inferred}
        if len(values) == 0:
            projected = None
        elif len(values) == 1:
            projected = values.pop()
        else:
            details = ", ".join(
                f"{n}: {'projected' if v else 'geographic'}" for n, v in inferred
            )
            warnings.warn(
                f"Inconsistent standard_names across coordinates for topology "
                f"'{topology}': {details}. Returning None.",
                UserWarning,
                stacklevel=2,
            )
            projected = None
        topology_dict[topology] = projected

    return topology_dict


@xr.register_dataset_accessor("ugrid_roles")
class UgridRolesAccessor:
    """
    Xarray Dataset "accessor" to retrieve the names of UGRID variables.

    Examples
    --------
    To get a list of the UGRID dummy variables in the dataset:

    >>> dataset.ugrid_roles.topology

    To get the names of the connectivity variables in the dataset:

    >>> dataset.ugrid_roles.connectivity

    Names can also be accessed directly through the topology:

    >>> dataset.ugrid_roles["mesh2d"]["node_dimension"]

    """

    def __init__(self, ds: xr.Dataset):
        self._ds = ds

    def __getitem__(self, key: str):
        if key not in self.topology:
            raise KeyError(key)
        return ChainMap(
            self.dimensions[key], self.coordinates[key], self.connectivity[key]
        )

    @property
    def topology(self) -> List[str]:
        """
        Get the names of the topology dummy variables, marked by a CF-role of
        ``mesh_topology``.

        Returns
        -------
        topology: List[str]
        """
        return _get_topology(self._ds)

    @property
    def coordinates(self) -> Dict[str, Dict[str, Tuple[List[str], List[str]]]]:
        """
        Get the names of the coordinate variables from the topology attributes.

        Returns a dictionary with the coordinates for the UGRID coordinates:

            * node coordinates
            * edge coordinates
            * face coordinates

        Multiple coordinates may be defined. The coordinates are grouped by
        their role (x or y).

        Returns
        -------
        coordinates: dict[str, dict[str, Tuple[List[str]]]]
        """
        return _get_coordinates(self._ds, self.topology)

    @property
    def dimensions(self) -> Dict[str, Dict[str, str]]:
        """
        Get the dimension names from the topology attributes and infer them
        from connectivity arrays or coordinates.

        Returns a dictionary with the UGRID dimensions per topology:

            * node dimension
            * edge dimension
            * face dimension

        Returns
        -------
        dimensions: dict[str, dict[str, str]]
        """
        return _get_dimensions(
            self._ds, self.topology, self.connectivity, self.coordinates
        )

    @property
    def connectivity(self) -> Dict[str, Dict[str, str]]:
        """
        Get the names of the variables containing the UGRID connectivity data.

            * face_node_connectivity
            * edge_node_connectivity
            * face_edge_connectivity
            * edge_face_connectivity

        Returns
        -------
        connectivity: Dict[str, Dict[str, str]]
        """
        return _get_connectivity(self._ds, self.topology)

    @property
    def grid_mapping_names(self) -> Dict[str, Optional[Set[str]]]:
        """
        Get the names of the grid mapping variables associated with each topology.

        Returns
        -------
        grid_mapping: dict[str, str | None]
        """
        return _get_grid_mapping_names(self._ds, self.topology, self.dimensions)

    @property
    def is_projected(self) -> Dict[str, bool | None]:
        """
        Infer whether each topology uses projected or geographic coordinates
        from the standard_name attributes of the coordinate variables.

        Returns
        -------
        is_projected: dict[str, bool | None]
            True if projected, False if geographic, None if indeterminate.
        """
        return _infer_projected(self._ds, self.topology, self.coordinates)

    def __repr__(self):
        dimensions = self.dimensions
        coordinates = self.coordinates
        connectivity = self.connectivity
        grid_mapping_names = self.grid_mapping_names
        is_projected = self.is_projected

        def make_text_section(subtitle, entries, vardict):
            tab = "    "
            rows = [f"{tab}{subtitle}"]
            for role in entries:
                if role in vardict:
                    rows += [f"{tab}{tab}{role}: {vardict[role]}"]
                else:
                    rows += [f"{tab}{tab}{role}: n/a"]
            rows.append("")
            return rows

        rows = []
        for topology in self.topology:
            topodim = self._ds[topology].attrs["topology_dimension"]
            rows += [f"UGRID {topodim}D Topology {topology}:"]
            rows += make_text_section(
                "Dimensions:", _DIM_NAMES[topodim], dimensions[topology]
            )
            rows += make_text_section(
                "Connectivity:", _CONNECTIVITY_NAMES[topodim], connectivity[topology]
            )
            rows += make_text_section(
                "Coordinates:", _COORD_NAMES[topodim], coordinates[topology]
            )

            # CRS summary line
            name = grid_mapping_names[topology]
            projected = is_projected[topology]
            if projected is True:
                crs_type = "projected"
            elif projected is False:
                crs_type = "geographic"
            else:
                crs_type = "unknown"
            name_str = name if name is not None else "n/a"
            rows += [
                f"    Coordinate Type: {crs_type}",
                f"Grid Mapping Name: {name_str}",
                "",
            ]

        return "\n".join(rows)



from typing import NamedTuple

import geopandas as gpd
import numpy as np
import pyproj
import pytest
import shapely
import xarray as xr
from matplotlib.collections import LineCollection
from scipy import sparse, spatial

import xugrid
from xugrid.ugrid.crs import CrsPlaceholder

from . import requires_meshkernel

try:
    import meshkernel as mk
except ImportError:
    pass

NAME = "network1d"


def grid1d(dataset=None, indexes=None, crs=None, attrs=None):
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 2.0],
        ]
    )
    grid = xugrid.Ugrid1d(
        node_x=xy[:, 0],
        node_y=xy[:, 1],
        fill_value=-1,
        edge_node_connectivity=np.array([[0, 1], [1, 2]]),
        dataset=dataset,
        indexes=indexes,
        crs=crs,
        attrs=attrs,
    )
    return grid


def test_ugrid1d_init():
    grid = grid1d()
    assert grid.name == NAME
    assert grid._dataset is None
    assert grid.node_x.flags["C_CONTIGUOUS"]
    assert grid.node_y.flags["C_CONTIGUOUS"]


def test_safe_attrs():
    # .attrs should return a copy
    grid = grid1d()
    assert grid.attrs == grid.attrs
    assert grid._attrs is not grid.attrs


def test_ugrid1d_alternative_init():
    custom_attrs = {
        "node_dimension": "nNetNode",
        "name": "mesh1d",
        "node_coordinates": "mesh1d_node_x mesh1d_node_y",
    }
    indexes = {"node_x": "mesh1d_node_x", "node_y": "mesh1d_node_y"}
    grid = grid1d(indexes=indexes, attrs=custom_attrs)
    assert grid.node_dimension == "nNetNode"
    assert grid.name == NAME
    # name in attrs should be overwritten by given name.
    assert grid._attrs["name"] == NAME

    with pytest.raises(ValueError, match="Provide either dataset or attrs, not both"):
        grid1d(dataset=xr.Dataset, attrs=custom_attrs)

    with pytest.raises(ValueError, match="indexes must be provided for dataset"):
        grid1d(dataset=xr.Dataset, indexes=None)

    with pytest.raises(ValueError, match="indexes must be provided for attrs"):
        grid1d(attrs=custom_attrs)


def test_ugrid1d_properties():
    # These are defined in the base class
    grid = grid1d()
    assert grid.node_dimension == f"{NAME}_nNodes"
    assert grid.edge_dimension == f"{NAME}_nEdges"
    assert grid.n_node == 3
    assert grid.n_edge == 2
    assert grid.facets == {"node": grid.node_dimension, "edge": grid.edge_dimension}
    expected_coords = np.array([[0.0, 0.0], [1.0, 1.0], [2.0, 2.0]])
    assert np.allclose(grid.node_coordinates, expected_coords)
    assert np.allclose(grid.edge_x, [0.5, 1.5])
    assert np.allclose(grid.edge_y, [0.5, 1.5])
    assert np.allclose(grid.edge_coordinates, np.column_stack([[0.5, 1.5], [0.5, 1.5]]))
    assert grid.bounds == (0.0, 0.0, 2.0, 2.0)
    assert isinstance(grid.node_edge_connectivity, sparse.csr_matrix)
    assert isinstance(grid.node_node_connectivity, sparse.csr_matrix)
    assert isinstance(grid.directed_node_node_connectivity, sparse.csr_matrix)
    assert isinstance(grid.directed_edge_edge_connectivity, sparse.csr_matrix)

    expected_coords = [
        [[0.0, 0.0], [1.0, 1.0]],
        [[1.0, 1.0], [2.0, 2.0]],
    ]
    actual_coords = grid.edge_node_coordinates
    assert actual_coords.shape == (2, 2, 2)
    assert np.allclose(actual_coords, expected_coords)
    assert isinstance(grid.attrs, dict)

    coords = grid.coords
    assert isinstance(coords, dict)
    assert np.array_equal(coords[grid.node_dimension], grid.node_coordinates)
    assert np.array_equal(coords[grid.edge_dimension], grid.edge_coordinates)

    with pytest.raises(ValueError, match="start_index must be 0 or 1, received: 2"):
        grid.start_index = 2
    grid.start_index = 1
    assert grid._start_index == 1

    assert isinstance(grid.node_kdtree, spatial.KDTree)
    assert isinstance(grid.edge_kdtree, spatial.KDTree)


def test_ugrid1d_optional_attributes():
    # Check whether edge coordinates end up in the _indexes
    ds = grid1d().to_dataset(optional_attributes=True)
    grid = xugrid.Ugrid1d.from_dataset(ds)
    assert grid._indexes == {
        "node_x": "network1d_node_x",
        "node_y": "network1d_node_y",
        "edge_x": "network1d_edge_x",
        "edge_y": "network1d_edge_y",
    }


def test_ugrid1d_egde_bounds():
    grid = grid1d()
    expected = np.array(
        [
            [0.0, 0.0, 1.0, 1.0],
            [1.0, 1.0, 2.0, 2.0],
        ]
    )
    actual = grid.edge_bounds
    assert actual.shape == (2, 4)
    assert np.allclose(actual, expected)


def test_validate_crs():
    grid = grid1d()
    assert grid._validate_crs(None, True) == (None, True)
    assert grid._validate_crs(None, False) == (None, False)

    placeholder = CrsPlaceholder({})
    assert grid._validate_crs(placeholder, True) == (placeholder, True)
    assert grid._validate_crs(placeholder, False) == (placeholder, False)

    crs = pyproj.CRS.from_epsg(28992)
    # CRS overrides projected argument
    expected = (crs, True)
    assert grid._validate_crs(crs, True) == expected
    assert grid._validate_crs(crs, False) == expected
    assert grid._validate_crs("EPSG:28992", True) == expected
    assert grid._validate_crs("EPSG:28992", False) == expected
    expected = (pyproj.CRS.from_epsg(4326), False)
    assert grid._validate_crs("EPSG:4326", True) == expected
    assert grid._validate_crs("EPSG:4326", False) == expected

    with pytest.raises(ValueError, match="Unsupported CRS"):
        # Test a geocentric non-projected, non-geographic CRS
        grid._validate_crs("EPSG:4328", True) == expected


def test_ugrid1d_update_coordinate_attrs():
    grid = grid1d()
    obj = xr.DataArray(np.ones(grid.n_edge), dims=(grid.edge_dimension,))
    obj = grid.assign_edge_coords(obj)
    grid._indexes["edge_x"] = "network1d_edge_x"
    grid._indexes["edge_y"] = "network1d_edge_y"
    grid.set_crs(epsg=4326)
    grid._update_coordinate_attrs(obj)
    assert obj["network1d_edge_x"].attrs["standard_name"] == "longitude"
    assert obj["network1d_edge_y"].attrs["standard_name"] == "latitude"


def test_set_crs():
    grid = grid1d()

    with pytest.raises(ValueError, match="Must pass either"):
        grid.set_crs()

    grid.set_crs("epsg:28992")
    assert grid.crs == pyproj.CRS.from_epsg(28992)

    # This is allowed: the same crs
    grid.set_crs("epsg:28992")
    assert grid.crs == pyproj.CRS.from_epsg(28992)

    # This is not allowed ...
    with pytest.raises(ValueError, match="The Ugrid already has a CRS"):
        grid.set_crs("epsg:4326")

    # Unless explicitly set with allow_override
    grid.set_crs("epsg:4326", allow_override=True)
    assert grid.crs == pyproj.CRS.from_epsg(4326)

    # Test espg alternative arg
    grid.crs = None
    grid.set_crs(epsg=28992)
    assert grid.crs == pyproj.CRS.from_epsg(28992)

    # Cannot set crs if a placeholder is in place.
    grid.crs = xugrid.ugrid.crs.CrsPlaceholder({})
    with pytest.raises(ValueError, match="The Ugrid already has a CRS"):
        grid.set_crs(epsg=28992)
    # Unless allow_override
    grid.set_crs(epsg=28992, allow_override=True)
    assert grid.crs == pyproj.CRS.from_epsg(28992)


def test_ugrid1d_assign_derived_coordinates():
    grid = grid1d()
    obj = xr.DataArray(np.ones(grid.n_edge), dims=(grid.edge_dimension,))
    obj = grid._assign_derived_coords(obj)
    assert "network1d_edge_x" in obj.coords
    assert "network1d_edge_y" in obj.coords


def test_to_crs():
    grid = grid1d()

    with pytest.raises(ValueError, match="Cannot transform naive geometries"):
        grid.to_crs("epsg:28992")

    grid.set_crs("epsg:4326")

    # Skip reprojection
    same = grid.to_crs("epsg:4326")
    assert np.allclose(same.node_coordinates, grid.node_coordinates)

    reprojected = grid.to_crs("epsg:28992")
    assert reprojected.crs == pyproj.CRS.from_epsg(28992)
    assert (~(grid.node_coordinates == reprojected.node_coordinates)).all()

    grid.crs = xugrid.ugrid.crs.CrsPlaceholder({})
    with pytest.raises(ValueError, match="Cannot transform geometries"):
        grid.to_crs(epsg=28992)


def test_ugrid1d_write_grid_mapping():
    grid = grid1d()
    grid.set_crs(epsg=28992)
    dataset = grid.to_dataset()
    assert "network1d_crs" in dataset
    assert dataset["network1d_crs"].attrs["name"] == "Amersfoort / RD New"


def test_to_dataset():
    def check_attrs(ds):
        attrs = ds[NAME].attrs.copy()
        attrs.pop("cf_role")
        attrs.pop("long_name")
        attrs.pop("topology_dimension")
        ds_contents = tuple(ds.dims) + tuple(ds.coords) + tuple(ds.data_vars)
        for values in attrs.values():
            # e.g node_coordinates are joined by a whitespace.
            for value in values.split(" "):
                assert value in ds_contents

    grid = grid1d()
    ds = grid.to_dataset()
    assert isinstance(ds, xr.Dataset)
    assert f"{NAME}" in ds
    assert f"{NAME}_nNodes" in ds.dims
    assert f"{NAME}_nEdges" in ds.dims
    assert f"{NAME}_node_x" in ds.coords
    assert f"{NAME}_node_y" in ds.coords
    assert f"{NAME}_edge_nodes" in ds
    check_attrs(ds)

    ds = grid.to_dataset(optional_attributes=True)
    assert f"{NAME}_edge_x" in ds.coords
    assert f"{NAME}_edge_y" in ds.coords
    check_attrs(ds)


def test_ugrid1d_dataset_roundtrip():
    grid = grid1d()
    ds = grid.to_dataset()
    grid2 = xugrid.Ugrid1d.from_dataset(grid.to_dataset())
    assert isinstance(grid2._dataset, xr.Dataset)
    assert grid2._dataset == ds


def test_ugrid1d_from_meshkernel():
    class Mesh1d(NamedTuple):
        node_x: np.ndarray
        node_y: np.ndarray
        edge_nodes: np.ndarray

    mesh1d = Mesh1d(
        node_x=np.array(
            [
                0.0,
                0.8975979,
                1.7951958,
                2.6927937,
                3.5903916,
                4.48798951,
                5.38558741,
                6.28318531,
            ]
        ),
        node_y=np.array(
            [
                0.00000000e00,
                7.81831482e-01,
                9.74927912e-01,
                4.33883739e-01,
                -4.33883739e-01,
                -9.74927912e-01,
                -7.81831482e-01,
                -2.44929360e-16,
            ]
        ),
        edge_nodes=np.array([0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 0, 0]),
    )

    grid = xugrid.Ugrid1d.from_meshkernel(mesh1d)
    assert grid.n_edge == 8
    assert np.allclose(mesh1d.node_x, grid.node_x)
    assert np.allclose(mesh1d.node_y, grid.node_y)
    assert np.allclose(grid.edge_node_connectivity, mesh1d.edge_nodes.reshape((8, 2)))


def test_clear_geometry_properties():
    grid = grid1d()
    for attr in [
        "_mesh",
        "_meshkernel",
        "_node_kdtree",
        "_edge_kdtree",
        "_celltree",
        "_xmin",
        "_xmax",
        "_ymin",
        "_ymax",
        "_edge_x",
        "_edge_y",
    ]:
        setattr(grid, attr, 1)
        grid._clear_geometry_properties()
        assert getattr(grid, attr) is None


def test_topology_dimension():
    grid = grid1d()
    assert grid.topology_dimension == 1


def test_dimensions():
    grid = grid1d()
    assert grid.node_dimension == f"{NAME}_nNodes"
    assert grid.edge_dimension == f"{NAME}_nEdges"
    assert grid.dims == {f"{NAME}_nNodes", f"{NAME}_nEdges"}
    assert grid.sizes == {
        f"{NAME}_nNodes": 3,
        f"{NAME}_nEdges": 2,
    }
    with pytest.warns(FutureWarning):
        assert grid.dimensions == grid.sizes


@requires_meshkernel
def test_mesh():
    grid = grid1d()
    assert isinstance(grid.mesh, mk.Mesh1d)


@requires_meshkernel
def test_meshkernel():
    grid = grid1d()
    assert isinstance(grid.meshkernel, mk.MeshKernel)


def test_is_cyclic():
    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 1.0, 0.0])
    edge_nodes = np.array(
        [
            [0, 1],
            [1, 2],
            [2, 0],
        ]
    )
    cycle = xugrid.Ugrid1d(x, y, -1, edge_nodes)
    assert cycle.is_cyclic

    edge_nodes = np.array(
        [
            [0, 1],
            [1, 2],
            [0, 2],
        ]
    )
    dag = xugrid.Ugrid1d(x, y, -1, edge_nodes)
    assert not dag.is_cyclic


def test_from_shapely():
    with pytest.raises(TypeError):
        xy = np.array(
            [
                [0.0, 0.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, 0.0],
            ]
        )
        xugrid.Ugrid1d.from_shapely(geometry=[shapely.polygons(xy)])

    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    grid = xugrid.Ugrid1d.from_shapely(geometry=shapely.linestrings(x, y))
    assert isinstance(grid, xugrid.Ugrid1d)


def test_from_geodataframe():
    with pytest.raises(TypeError, match="Expected GeoDataFrame"):
        xugrid.Ugrid1d.from_geodataframe(1)

    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    gdf = gpd.GeoDataFrame(geometry=[shapely.linestrings(x, y)])
    grid = xugrid.Ugrid1d.from_geodataframe(gdf)
    assert isinstance(grid, xugrid.Ugrid1d)


def test_to_shapely():
    grid = grid1d()

    points = grid.to_shapely(f"{NAME}_nNodes")
    assert isinstance(points[0], shapely.Geometry)

    lines = grid.to_shapely(f"{NAME}_nEdges")
    assert isinstance(lines[0], shapely.Geometry)


def test_sel():
    grid = grid1d()
    obj = xr.DataArray(
        data=[0, 1],
        dims=[grid.edge_dimension],
    )
    with pytest.raises(ValueError, match="Ugrid1d only supports slice indexing"):
        grid.sel(obj=obj, x=1.0, y=1.0)
    with pytest.raises(ValueError, match="Ugrid1d does not support steps"):
        grid.sel(obj=obj, x=slice(0, 2, 1), y=slice(0, 2, 1))
    with pytest.raises(ValueError, match="slice start should be smaller"):
        grid.sel(obj, x=slice(2, 0), y=slice(0, 2))

    actual = grid.sel(obj=obj, x=slice(0, 1), y=slice(0, 1))
    assert isinstance(actual, tuple)
    new_obj, new_grid = actual

    assert isinstance(new_obj, xr.DataArray)
    assert isinstance(new_grid, xugrid.Ugrid1d)
    assert new_obj.dims[0] == f"{NAME}_nEdges"
    assert new_grid.edge_dimension == f"{NAME}_nEdges"
    assert np.array_equal(new_obj.values, [0])


def test_sel_points():
    grid = grid1d()
    obj = xr.DataArray(
        data=[10, 11],
        dims=[grid.edge_dimension],
    )
    x = [1.5, 0.5, 0.0]
    y = [1.5, 0.5, 0.1]

    actual = grid.sel_points(obj=obj, x=x, y=y, out_of_bounds="ignore")
    np.testing.assert_allclose(actual.values, [11, 10, np.nan])
    np.testing.assert_allclose(actual["network1d_x"].values, x)
    np.testing.assert_allclose(actual["network1d_y"].values, y)

    # Check if out_of_bounds raises ValueError
    with pytest.raises(ValueError):
        grid.sel_points(obj=obj, x=x, y=y, out_of_bounds="raise")

    # Test with tolerance
    actual = grid.sel_points(obj=obj, x=x, y=y, tolerance=0.1, out_of_bounds="ignore")
    np.testing.assert_allclose(actual.values, [11, 10, 10])

    obj = xr.DataArray(
        data=np.arange(grid.n_node),
        dims=[grid.node_dimension],
    )
    actual = grid.sel_points(obj=obj, x=x, y=y, out_of_bounds="ignore")
    np.testing.assert_allclose(actual.values, [1, 0, np.nan])


def test_intersect_line():
    grid = grid1d()
    obj = xr.DataArray([0, 1, 2, 3], dims=[grid.edge_dimension])

    p0 = (2.0, 0.0)
    p1 = (0.0, 1.0)
    actual = grid.intersect_line(obj, start=p0, end=p1)
    expected_s = np.sqrt((2 / 3) ** 2 + (1 + 1 / 3) ** 2)
    assert isinstance(actual, xr.DataArray)
    assert actual.dims == (grid.edge_dimension,)
    assert np.array_equal(actual.to_numpy(), [0])
    assert np.allclose(actual[f"{NAME}_x"], [(2 / 3)])
    assert np.allclose(actual[f"{NAME}_y"], [(2 / 3)])
    assert np.allclose(actual[f"{NAME}_s"], [expected_s])

    actual = grid.intersect_line(obj, start=p1, end=p0)
    assert np.array_equal(actual.to_numpy(), [0])


def test_intersect_linestring():
    grid = grid1d()
    obj = xr.DataArray([0, 1, 2], dims=[grid.edge_dimension])
    linestring = shapely.geometry.LineString(
        [
            [0.0, 2.0],
            [1.0, 0.0],
            [1.5, 2.0],
        ]
    )
    actual = grid.intersect_linestring(obj, linestring)
    assert isinstance(actual, xr.DataArray)
    assert actual.dims == (grid.edge_dimension,)
    assert np.array_equal(actual.to_numpy(), [0, 1])
    assert np.allclose(actual[f"{NAME}_x"], [(2 / 3), 1 + (1 / 3)])
    assert np.allclose(actual[f"{NAME}_y"], [(2 / 3), 1 + (1 / 3)])
    assert np.allclose(actual[f"{NAME}_s"], [1.491, 3.610], atol=1e-3)


def test_topology_subset():
    grid = grid1d()
    edge_indices = np.array([1])
    actual = grid.topology_subset(edge_indices)
    assert np.array_equal(actual.edge_node_connectivity, [[0, 1]])
    assert np.array_equal(actual.node_x, [1.0, 2.0])
    assert np.array_equal(actual.node_y, [1.0, 2.0])


def test_reindex_like():
    grid = grid1d()
    index = np.array([1, 0])
    reordered = grid.topology_subset(index)
    obj = xr.DataArray(index, dims=(reordered.edge_dimension))
    reindexed = reordered.reindex_like(grid, obj=obj)
    assert np.array_equal(reindexed, [0, 1])


def test_ugrid1d_plot():
    grid = grid1d()
    primitive = grid.plot()
    assert isinstance(primitive, LineCollection)


def test_ugrid1d_rename():
    grid = grid1d()
    original_indexes = grid._indexes.copy()
    original_attrs = grid._attrs.copy()

    renamed = grid.rename("__renamed")

    # Check that original is unchanged
    assert grid._attrs == original_attrs
    assert grid._indexes == original_indexes
    assert renamed._attrs == {
        "cf_role": "mesh_topology",
        "long_name": "Topology data of 1D network",
        "topology_dimension": 1,
        "node_dimension": "__renamed_nNodes",
        "edge_dimension": "__renamed_nEdges",
        "edge_node_connectivity": "__renamed_edge_nodes",
        "node_coordinates": "__renamed_node_x __renamed_node_y",
        "edge_coordinates": "__renamed_edge_x __renamed_edge_y",
    }
    assert renamed._indexes == {
        "node_x": "__renamed_node_x",
        "node_y": "__renamed_node_y",
    }
    assert renamed.name == "__renamed"


def test_ugrid1d_rename_with_dataset():
    grid = grid1d()
    grid2 = xugrid.Ugrid1d.from_dataset(grid.to_dataset())
    original_dataset = grid2._dataset.copy()

    renamed2 = grid2.rename("__renamed")
    dataset = renamed2._dataset
    assert grid2._dataset.equals(original_dataset)
    assert sorted(dataset.data_vars) == ["__renamed", "__renamed_edge_nodes"]
    assert sorted(dataset.dims) == ["__renamed_nEdges", "__renamed_nNodes", "two"]
    assert sorted(dataset.coords) == ["__renamed_node_x", "__renamed_node_y"]


def test_topology_sort_by_dfs():
    grid = grid1d()
    vertices = grid.topological_sort_by_dfs()
    assert isinstance(vertices, np.ndarray)
    assert np.array_equal(vertices, [0, 1, 2])


def test_contract_vertices():
    grid = grid1d()
    new = grid.contract_vertices(indices=[0, 2])
    assert isinstance(new, xugrid.Ugrid1d)
    assert new.n_node == 2
    # The nodes have been renumbered
    assert np.array_equal(new.edge_node_connectivity, [[0, 1]])


def test_remove_self_loops():
    xy = np.array(
        [
            [0.0, 10.0],
            [3.0, 13.0],  # Disconnected
            [1.0, 11.0],
            [2.0, 12.0],
        ]
    )
    edges = np.array(
        [
            [0, 0],
            [1, 1],
            [0, 2],
            [2, 2],
            [2, 3],
            [3, 3],
        ]
    )
    grid = xugrid.Ugrid1d(*xy.T, -1, edges)
    actual = grid.remove_self_loops()
    assert isinstance(actual, xugrid.Ugrid1d)
    assert actual.n_node == 3
    assert actual.n_edge == 2
    assert np.array_equal(actual.node_x, [0.0, 1.0, 2.0])
    assert np.array_equal(actual.node_y, [10.0, 11.0, 12.0])
    assert np.array_equal(actual.edge_node_connectivity, [[0, 1], [1, 2]])


def test_get_connectivity_matrix():
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [3.0, 0.0],
        ]
    )
    grid = xugrid.Ugrid1d(
        node_x=xy[:, 0],
        node_y=xy[:, 1],
        fill_value=-1,
        edge_node_connectivity=np.array([[0, 1], [1, 2]]),
    )
    with pytest.raises(ValueError, match="Expected network1d_nNodes; got: abc"):
        grid.get_connectivity_matrix(dim="abc", xy_weights=True)

    connectivity = grid.get_connectivity_matrix(grid.node_dimension, True)
    assert isinstance(connectivity, sparse.csr_matrix)
    assert np.allclose(connectivity.data, [1.5, 1.5, 0.75, 0.75])
    assert np.array_equal(connectivity.indices, [1, 0, 2, 1])


def test_get_coordinates():
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [3.0, 0.0],
        ]
    )
    grid = xugrid.Ugrid1d(
        node_x=xy[:, 0],
        node_y=xy[:, 1],
        fill_value=-1,
        edge_node_connectivity=np.array([[0, 1], [1, 2]]),
    )
    with pytest.raises(
        ValueError, match="Expected network1d_nNodes or network1d_nEdges; got: abc"
    ):
        grid.get_coordinates(dim="abc")

    assert isinstance(grid.get_coordinates(grid.node_dimension), np.ndarray)
    assert isinstance(grid.get_coordinates(grid.edge_dimension), np.ndarray)


def test_equals():
    grid = grid1d()
    grid_copy = grid1d()
    assert grid.equals(grid)
    assert grid.equals(grid_copy)
    xr_grid = grid.to_dataset()
    assert not grid.equals(xr_grid)
    grid_copy._attrs["attr"] = "something_else"
    # Dataset.identical is called so returns False
    assert not grid.equals(grid_copy)


def test_ugrid1d_create_data_array():
    grid = grid1d()

    uda = grid.create_data_array(np.zeros(grid.n_node), facet="node")
    assert isinstance(uda, xugrid.UgridDataArray)

    uda = grid.create_data_array(np.zeros(grid.n_edge), facet="edge")
    assert isinstance(uda, xugrid.UgridDataArray)

    # Error on facet
    with pytest.raises(ValueError, match="Invalid facet"):
        grid.create_data_array([1, 2, 3], facet="face")

    # Error on on dimensions
    with pytest.raises(ValueError, match="Can only create DataArrays from 1D arrays"):
        grid.create_data_array([[1, 2, 3]], facet="node")

    # Error on size
    with pytest.raises(ValueError, match="Conflicting sizes"):
        grid.create_data_array([1, 2, 3, 4], facet="node")


def test_ugrid1d_format_connectivity():
    grid = grid1d()
    assert isinstance(grid.edge_node_connectivity, np.ndarray)
    assert isinstance(
        grid.format_connectivity_as_sparse(grid.edge_node_connectivity),
        sparse.csr_matrix,
    )
    assert isinstance(grid.node_node_connectivity, sparse.csr_matrix)
    assert isinstance(
        grid.format_connectivity_as_dense(grid.node_node_connectivity), np.ndarray
    )
    assert isinstance(
        grid.format_connectivity_as_sparse(grid.node_node_connectivity.tocoo()),
        sparse.csr_matrix,
    )


def test_ugrid1d_length():
    grid = grid1d()
    length = grid.edge_length
    assert isinstance(length, np.ndarray)
    assert length.shape == (grid.n_edge,)
    assert np.allclose(length, [np.sqrt(2), np.sqrt(2)])


def test_ugrid1d_refine_by_vertices():
    node_xy = np.array(
        [
            [0.0, 0.0],
            [5.0, 5.0],
            [10.0, 5.0],
            [15.0, 0.0],
            [15.0, 10.0],
        ]
    )
    edge_nodes = np.array(
        [
            [0, 1],
            [1, 2],
            [2, 3],
            [2, 4],
        ]
    )
    grid = xugrid.Ugrid1d(*node_xy.T, -1, edge_nodes)
    vertices = np.array(
        [
            [7.5, 5.0],
            [12.5, 2.5],
            [12.5, 7.5],
            [1.0, 1.0],
            [4.0, 4.0],
        ]
    )
    expected_edge_node_coordinates = np.array(
        [
            [[0.0, 0.0], [1.0, 1.0]],
            [[1.0, 1.0], [4.0, 4.0]],
            [[4.0, 4.0], [5.0, 5.0]],
            [[5.0, 5.0], [7.5, 5.0]],
            [[7.5, 5.0], [10.0, 5.0]],
            [[10.0, 5.0], [12.5, 2.5]],
            [[12.5, 2.5], [15.0, 0.0]],
            [[10.0, 5.0], [12.5, 7.5]],
            [[12.5, 7.5], [15.0, 10.0]],
        ]
    )
    expected_edge_node_connectivity = np.array(
        [[0, 8], [8, 9], [9, 1], [1, 5], [5, 2], [2, 6], [6, 3], [2, 7], [7, 4]]
    )
    expected_new_index = np.array([8, 9, 5, 6, 7])
    new = grid.refine_by_vertices(vertices)
    np.testing.assert_allclose(
        new.edge_node_coordinates, expected_edge_node_coordinates
    )
    np.testing.assert_equal(new.edge_node_connectivity, expected_edge_node_connectivity)
    # Test with return_index option = True
    new, new_index = grid.refine_by_vertices(vertices, return_index=True)
    np.testing.assert_allclose(
        new.edge_node_coordinates, expected_edge_node_coordinates
    )
    np.testing.assert_equal(new_index, expected_new_index)
    actual_vertices = new.node_coordinates[new_index]
    # Test if returned indices indeed point to inserted vertices (regardless of order)
    np.testing.assert_array_almost_equal(
        np.sort(actual_vertices.flat), np.sort(vertices.flat)
    )

    # Test if error upon trying to insert vertices that are not present
    vertices_wrong = np.array(
        [
            [5.0, 6.0],
            [12.5, 2.5],
            [12.5, 7.5],
        ]
    )
    with pytest.raises(
        ValueError, match="The following vertices are not located on any edge"
    ):
        grid.refine_by_vertices(vertices_wrong)

    # Test tolerance passed through correctly
    vertices[:, 0] += 0.01
    expected_edge_node_coordinates = np.array(
        [
            [[0.0, 0.0], [1.01, 1.0]],
            [[1.01, 1.0], [4.01, 4.0]],
            [[4.01, 4.0], [5.0, 5.0]],
            [[5.0, 5.0], [7.51, 5.0]],
            [[7.51, 5.0], [10.0, 5.0]],
            [[10.0, 5.0], [12.51, 2.5]],
            [[12.51, 2.5], [15.0, 0.0]],
            [[10.0, 5.0], [12.51, 7.5]],
            [[12.51, 7.5], [15.0, 10.0]],
        ]
    )
    new = grid.refine_by_vertices(vertices, tolerance=0.01)
    np.testing.assert_allclose(
        new.edge_node_coordinates, expected_edge_node_coordinates
    )
    np.testing.assert_equal(new.edge_node_connectivity, expected_edge_node_connectivity)


def test_nearest_interpolate():
    node_x = np.array([0.0, 1.0, 1.9, 3.0, 4.0])
    node_y = np.zeros_like(node_x)
    edge_node_connectivity = np.array(
        [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 4],
        ]
    )
    grid = xugrid.Ugrid1d(node_x, node_y, -1, edge_node_connectivity)
    data = np.array([0.0, np.nan, np.nan, np.nan, 4.0])
    nodedim = grid.node_dimension
    actual = grid._nearest_interpolate(data, nodedim, np.inf)
    assert np.allclose(actual, np.array([0.0, 0.0, 0.0, 4.0, 4.0]))

    actual = grid._nearest_interpolate(data, nodedim, 1.1)
    assert np.allclose(actual, np.array([0.0, 0.0, np.nan, 4.0, 4.0]), equal_nan=True)

    with pytest.raises(ValueError, match="All values are NA."):
        grid._nearest_interpolate(np.full_like(data, np.nan), nodedim, np.inf)

    # Test for edge dimension
    data = np.array([0.0, np.nan, np.nan, 4.0])
    actual = grid._nearest_interpolate(data, grid.edge_dimension, np.inf)
    assert np.allclose(actual, np.array([0.0, 0.0, 4.0, 4.0]))


def test_locate_nearest():
    grid = grid1d()
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 2.0],
        ]
    )
    indices = grid.locate_nearest_node(xy)
    assert np.array_equal(indices, [0, 1, 2])

    indices = grid.locate_nearest_edge(grid.edge_coordinates)
    assert np.array_equal(indices, [0, 1])

    indices = grid.locate_nearest_node([[0.0, 10.0]], max_distance=1.0)
    assert np.array_equal(indices, [-1])

    indices = grid.locate_nearest_edge([[0.0, 10.0]], max_distance=1.0)
    assert np.array_equal(indices, [-1])



import numpy as np
import pytest
import xarray as xr

from xugrid.regrid.structured import StructuredGrid1d, StructuredGrid2d

# Testgrids
# --------
# grid a(x):               |______50_____|_____100_____|_____150_____|               -> source
# grid b(x):        |______25_____|______75_____|_____125_____|_____175_____|        -> target
# --------
# grid c(x):            |______40_____|______90_____|_____140_____|____190_____|     -> target
# --------
# grid d(x):              |__30__|__55__|__80_|__105__|                              -> target
# --------
# grid e(x):              |__30__|____67.5____|__105__|                              -> target
# --------


def test_init_1d(grid_data_a_1d):
    assert isinstance(grid_data_a_1d, StructuredGrid1d)
    with pytest.raises(TypeError):
        StructuredGrid1d(1)


def test_init_2d(grid_data_a_2d):
    assert isinstance(grid_data_a_2d, StructuredGrid2d)
    with pytest.raises(TypeError):
        StructuredGrid2d(1)


def assert_expected_overlap(
    actual_source,
    actual_target,
    actual_weights,
    expected_source,
    expected_target,
    expected_weights,
):
    # Numpy 2.0 release has change sorting behavior of non-stable sorting:
    # https://numpy.org/doc/stable/release/2.0.0-notes.html#minor-changes-in-behavior-of-sorting-functions
    # So the comparison method must be robust to work for numpy <2.0 and >=2.0.
    actual_mapping = np.column_stack((actual_target, actual_source))
    expected_mapping = np.column_stack((expected_target, expected_source))
    actual, actual_sorter = np.unique(actual_mapping, axis=0, return_index=True)
    expected, expected_sorter = np.unique(expected_mapping, axis=0, return_index=True)
    assert np.array_equal(actual, expected)
    assert np.allclose(actual_weights[actual_sorter], expected_weights[expected_sorter])


def test_overlap_1d(
    grid_data_a_1d, grid_data_b_1d, grid_data_b_flipped_1d, grid_data_e_1d
):
    # --------
    # source   targets  weight
    # node 0   0, 1     25 m
    # node 1   1, 2     25 m
    # node 2   2, 3     25 m
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.overlap(grid_data_b_1d, relative=False),
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([0, 1, 1, 2, 2, 3]),
        np.array([25, 25, 25, 25, 25, 25]),
    )

    # flipped axis (y-axis)
    # --------
    # source   targets  weight
    # 0        0, 1     25 m
    # 1        1, 2     25 m
    # 2        2, 3     25 m
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.overlap(grid_data_b_flipped_1d, relative=False),
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([2, 3, 1, 2, 0, 1]),
        np.array([25, 25, 25, 25, 25, 25]),
    )

    # non-equidistant
    # --------
    # source   targets  weight
    # node 0   0, 1     17.5 m, 32.5 m
    # node 1   1, 2     17.5 m, 25.0 m
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.overlap(grid_data_e_1d, relative=False),
        np.array([0, 0, 1, 1]),
        np.array([0, 1, 1, 2]),
        np.array([17.5, 32.5, 17.5, 25.0]),
    )

    # relative
    # --------
    # source   targets  weight
    # node 0   0, 1     17.5 m, 32.5 m
    # node 1   1, 2     17.5 m, 25.0 m
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.overlap(grid_data_e_1d, relative=True),
        np.array([0, 0, 1, 1]),
        np.array([0, 1, 1, 2]),
        np.array([17.5 / 50.0, 32.5 / 50.0, 17.5 / 50.0, 25.0 / 50.0]),
    )


def test_overlap_2d(grid_data_a_2d, grid_data_b_2d):
    # --------
    # source   targets            weights
    # 0        0,   1,  4,  5     625 m
    # 1        1,   2,  5,  6     625 m
    # 2        2,   3,  6,  7     625 m
    # 3        4,   5,  8,  9     625 m
    # 4        5,   6,  9, 10     625 m
    # 5        6,   7, 10, 11     625 m
    # 6        8,   9, 12, 13     625 m
    # 7        9,  10, 13, 14     625 m
    # 8        10, 11, 14, 15     625 m
    # --------
    assert_expected_overlap(
        *grid_data_a_2d.overlap(grid_data_b_2d, relative=False),
        expected_source=np.array(
            [
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                2,
                2,
                2,
                2,
                3,
                3,
                3,
                3,
                4,
                4,
                4,
                4,
                5,
                5,
                5,
                5,
                6,
                6,
                6,
                6,
                7,
                7,
                7,
                7,
                8,
                8,
                8,
                8,
            ]
        ),
        expected_target=np.array(
            [
                0,
                4,
                5,
                1,
                2,
                6,
                5,
                1,
                2,
                3,
                7,
                6,
                8,
                9,
                5,
                4,
                9,
                5,
                10,
                6,
                10,
                11,
                7,
                6,
                9,
                8,
                12,
                13,
                10,
                14,
                13,
                9,
                10,
                11,
                14,
                15,
            ]
        ),
        expected_weights=np.full(36, 625.0),
    )


def test_locate_centroids_1d(
    grid_data_a_1d, grid_data_b_1d, grid_data_b_flipped_1d, grid_data_e_1d
):
    # --------
    # source   target  weight
    # 0        1       1
    # 1        2       1
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.locate_centroids(grid_data_b_1d),
        np.array([0, 1]),
        np.array([1, 2]),
        np.ones(2),
    )

    # flipped axis (y-axis)
    # --------
    # source   target  weight
    # 0        1       1
    # 1        2       1
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.locate_centroids(grid_data_b_flipped_1d),
        np.array([0, 1]),
        np.array([2, 1]),
        np.ones(2),
    )

    # non-equidistant
    # --------
    # source   target  weight
    # 0        0, 1    1, 1
    # 1        2       1
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.locate_centroids(grid_data_e_1d),
        np.array([0, 0, 1]),
        np.array([0, 1, 2]),
        np.ones(3),
    )


def test_locate_centroids_2d(grid_data_a_2d, grid_data_b_2d):
    # --------
    # source   target  weight
    # 0        5       1
    # 1        6       1
    # 3        9       1
    # 4        10      1
    # --------
    assert_expected_overlap(
        *grid_data_a_2d.locate_centroids(grid_data_b_2d, None),
        np.array([0, 1, 3, 4]),
        np.array([5, 6, 9, 10]),
        np.ones(4),
    )


def test_linear_weights_1d(
    grid_data_a_1d,
    grid_data_b_1d,
    grid_data_b_flipped_1d,
    grid_data_c_1d,
    grid_data_d_1d,
    grid_data_e_1d,
):
    # --------
    # source   target  weight
    # 0   ->   1       50%
    # 1   ->   1       50%
    # 1   ->   2       50%
    # 2   ->   2       50%
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.linear_weights(grid_data_b_1d),
        np.array([0, 1, 1, 2]),
        np.array([1, 1, 2, 2]),
        np.array([0.5, 0.5, 0.5, 0.5]),
    )

    # flipped axis (y-axis)
    # --------
    # source   target  weight
    # 0   ->   2       50%
    # 1   ->   2       50%
    # 1   ->   1       50%
    # 2   ->   1       50%
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.linear_weights(grid_data_b_flipped_1d),
        np.array([2, 1, 1, 0]),
        np.array([1, 1, 2, 2]),
        np.array([0.5, 0.5, 0.5, 0.5]),
    )

    # --------
    # source   target  weight
    # 1        1       80%
    # 0        1       20%
    # 2        2       80%
    # 1        2       20%
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.linear_weights(grid_data_c_1d),
        np.array([0, 0, 1, 0, 2, 1]),
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([0.0, 1.0, 0.8, 0.2, 0.8, 0.2]),
    )

    # --------
    # source   target  weight
    # 0        1       90%
    # 1        1       10%
    # 0        2       40%
    # 1        2       60% *reversed in output
    # 1        3       90%
    # 2        3       10%
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.linear_weights(grid_data_d_1d),
        np.array([0, 0, 0, 1, 1, 0, 1, 2]),
        np.array([0, 0, 1, 1, 2, 2, 3, 3]),
        np.array([0.0, 0.1, 0.9, 0.1, 0.6, 0.4, 0.9, 0.1]),
    )

    # non-equidistant
    # --------
    # source   target  weight
    # 0        1       65%
    # 1        1       35%
    # 1        2       90%
    # 2        2       10%
    # --------
    assert_expected_overlap(
        *grid_data_a_1d.linear_weights(grid_data_e_1d),
        np.array([0, 0, 0, 1, 1, 2]),
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([0.0, 1.0, 0.65, 0.35, 0.9, 0.1]),
    )

    # 1-1 grid
    # --------
    # source   target  weight
    # 1        1       100%
    # 0        1       0%
    # 2        2       100%
    # 1        2       0%
    # --------
    assert_expected_overlap(
        *grid_data_b_1d.linear_weights(grid_data_b_1d),
        np.array([0, 0, 1, 0, 2, 1, 3, 2]),
        np.array([0, 0, 1, 1, 2, 2, 3, 3]),
        np.array([0.0, 1.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0]),
    )


def test_linear_weights_2d(
    grid_data_a_2d, grid_data_a_layered_2d, grid_data_b_2d, grid_data_c_2d
):
    # --------
    # source   targets     weight
    # 5        0, 1, 3, 4  25%
    # 6        1, 2, 4, 5  25%
    # 9        3, 4, 6, 7  25%
    # 10       4, 5, 7, 8  25%
    # --------
    assert_expected_overlap(
        *grid_data_a_2d.linear_weights(grid_data_b_2d),
        np.array([3, 4, 1, 0, 5, 4, 1, 2, 6, 7, 4, 3, 8, 7, 4, 5]),
        np.array([5, 5, 5, 5, 6, 6, 6, 6, 9, 9, 9, 9, 10, 10, 10, 10]),
        np.array([0.25] * 16),
    )

    # --------
    # source   targets      weight
    # 4        0, 0, 3, 3   0%  50% 0%  50%
    # 5        0, 1, 3, 4   10%	40%	10%	40%
    # 6        1, 2, 4, 5   10%	40%	10%	40%
    # 8        3, 3, 6, 6   0%  50% 0%  50%
    # 9        3, 4, 6, 7   10%	40%	10%	40%
    # 10       4, 5, 7, 8   10%	40%	10%	40%
    # --------
    assert_expected_overlap(
        *grid_data_a_layered_2d.linear_weights(grid_data_c_2d),
        np.array(
            [0, 0, 3, 3, 1, 0, 3, 4, 5, 4, 2, 1, 3, 3, 6, 6, 4, 3, 7, 6, 8, 5, 4, 7]
        ),
        np.array(
            [4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 8, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10]
        ),
        np.array(
            [
                0.0,
                0.5,
                0.0,
                0.5,
                0.4,
                0.1,
                0.1,
                0.4,
                0.4,
                0.1,
                0.4,
                0.1,
                0.0,
                0.5,
                0.0,
                0.5,
                0.4,
                0.1,
                0.4,
                0.1,
                0.4,
                0.4,
                0.1,
                0.1,
            ]
        ),
    )

    # 1-1
    # --------
    # source   targets      weight
    # 0-15     0-15         0% 0% 0% 100% (or shuffled)
    # result should be 1:1 mapping
    # --------
    source, target, weights = grid_data_b_2d.linear_weights(grid_data_b_2d)
    expected_target = np.repeat(np.arange(16), 4)
    assert np.array_equal(target, expected_target)
    assert np.array_equal(np.unique(weights), [0, 1])
    check_source = source[weights != 0]
    assert np.array_equal(check_source, np.arange(16))


def test_nonscalar_dx():
    da = xr.DataArray(
        [1, 2, 3], coords={"x": [1, 2, 3], "dx": ("x", [1, 1, 1])}, dims=("x",)
    )
    grid = StructuredGrid1d(da, name="x")
    actual = xr.DataArray([1, 2, 3], coords=grid.coords, dims=grid.dims)
    assert actual.identical(da)


def test_directional_bounds():
    da = xr.DataArray([1, 2, 3], coords={"y": [1, 2, 3]}, dims=("y",))
    decreasing = da.isel(y=slice(None, None, -1))
    grid_inc = StructuredGrid1d(da, name="y")
    grid_dec = StructuredGrid1d(decreasing, name="y")
    assert grid_inc.flipped is False
    assert grid_dec.flipped is True
    assert np.array_equal(grid_inc.bounds, grid_dec.bounds)
    assert np.array_equal(
        grid_inc.directional_bounds, grid_dec.directional_bounds[::-1]
    )



import dask.array as da
import numpy as np
import pytest
import xarray as xr

import xugrid as xu
from xugrid.regrid.structured import StructuredGrid1d, StructuredGrid2d

# Testgrids
# --------
# grid a(x):               |______50_____|_____100_____|_____150_____|               -> source
# grid b(x):        |______25_____|______75_____|_____125_____|_____175_____|        -> target
# --------
# grid c(x):            |______40_____|______90_____|_____140_____|____190_____|     -> target
# --------
# grid d(x):              |__30__|__55__|__80_|__105__|                              -> target
# --------
# grid e(x):              |__30__|_____67.5___|__105__|                              -> target
# --------


@pytest.fixture(scope="function")
def disk():
    return xu.data.disk()["face_z"]


@pytest.fixture(scope="function")
def disk_layered(disk):
    layer = xr.DataArray([1.0, 2.0, 3.0], coords={"layer": [1, 2, 3]}, dims=("layer",))
    # Disk is first in multiplication, to ensure that object is promoted to UgridDataArray
    disk_layered = disk * layer
    return disk_layered.transpose("layer", disk.ugrid.grid.face_dimension)


@pytest.fixture(scope="function")
def quads_0_25():
    dx = 0.25
    xmin, ymin, xmax, ymax = xu.data.disk().ugrid.total_bounds
    x = np.arange(xmin, xmax, dx) + 0.5 * dx
    y = np.arange(ymin, ymax, dx) + 0.5 * dx

    da = xr.DataArray(
        data=np.full((y.size, x.size), np.nan),
        coords={"y": y, "x": x},
        dims=[
            "y",
            "x",
        ],
    )
    return xu.UgridDataArray.from_structured2d(da)


@pytest.fixture(scope="function")
def quads_structured():
    dx = 1.0
    xmin, ymin, xmax, ymax = xu.data.disk().ugrid.total_bounds
    x = np.arange(xmin, xmax, dx) + 0.5 * dx
    y = np.arange(ymin, ymax, dx) + 0.5 * dx

    da = xr.DataArray(
        data=np.full((y.size, x.size), 1.0),
        coords={"y": y, "x": x},
        dims=[
            "y",
            "x",
        ],
    )
    return da


@pytest.fixture(scope="function")
def quads_1():
    dx = 1.0
    xmin, ymin, xmax, ymax = xu.data.disk().ugrid.total_bounds
    x = np.arange(xmin, xmax, dx) + 0.5 * dx
    y = np.arange(ymin, ymax, dx) + 0.5 * dx

    da = xr.DataArray(
        data=np.full((y.size, x.size), np.nan),
        coords={"y": y, "x": x},
        dims=[
            "y",
            "x",
        ],
    )
    return xu.UgridDataArray.from_structured2d(da)


@pytest.fixture(scope="function")
def grid_data_a():
    return xr.DataArray(
        data=np.arange(9).reshape((3, 3)),
        dims=["y", "x"],
        coords={
            "y": np.array([150, 100, 50]),
            "x": np.array([50, 100, 150]),
            "dx": 50.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_a_layered():
    return xr.DataArray(
        data=np.arange(18).reshape((2, 3, 3)),
        dims=["layer", "y", "x"],
        coords={
            "layer": np.arange(2) + 1,
            "y": np.array([150, 100, 50]),
            "x": np.array([50, 100, 150]),
            "dx": 50.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_b():
    return xr.DataArray(
        data=np.zeros(16).reshape((4, 4)),
        dims=["y", "x"],
        coords={
            "y": np.array([175, 125, 75, 25]),
            "x": np.array([25, 75, 125, 175]),
            "dx": 50.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_c():
    return xr.DataArray(
        data=np.arange(16).reshape((4, 4)),
        dims=["y", "x"],
        coords={
            "y": np.array([175, 125, 75, 25]),
            "x": np.array([40, 90, 140, 190]),
            "dx": 50.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_d():
    return xr.DataArray(
        data=np.arange(16).reshape((4, 4)),
        dims=["y", "x"],
        coords={
            "y": np.array([175, 125, 75, 25]),
            "x": np.array([30, 55, 80, 105]),
            "dx": 25.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_e():
    return xr.DataArray(
        data=np.zeros((4, 3, 2)),
        dims=["y", "x", "nbounds"],
        coords={
            "y": np.array([175, 125, 75, 25]),
            "x": np.array([30, 67.5, 105]),
            "dx": 25,
            "dy": -50.0,
            "xbounds": (
                ["x", "nbounds"],
                np.column_stack(
                    (np.array([17.5, 42.5, 92.5]), np.array([42.5, 92.5, 117.5]))
                ),
            ),
            "nbounds": np.arange(2),
        },
    )


@pytest.fixture(scope="function")
def grid_data_a_1d(grid_data_a):
    return StructuredGrid1d(grid_data_a, "x")


@pytest.fixture(scope="function")
def grid_data_a_layered_1d(grid_data_a_layered):
    return StructuredGrid1d(grid_data_a_layered, "x")


@pytest.fixture(scope="function")
def grid_data_a_2d(grid_data_a):
    return StructuredGrid2d(grid_data_a, "x", "y")


@pytest.fixture(scope="function")
def grid_data_a_layered_2d(grid_data_a_layered):
    return StructuredGrid2d(grid_data_a_layered, "x", "y")


@pytest.fixture(scope="function")
def grid_data_b_1d(grid_data_b):
    return StructuredGrid1d(grid_data_b, "x")


@pytest.fixture(scope="function")
def grid_data_b_flipped_1d(grid_data_b):
    return StructuredGrid1d(grid_data_b, "y")


@pytest.fixture(scope="function")
def grid_data_c_1d(grid_data_c):
    return StructuredGrid1d(grid_data_c, "x")


@pytest.fixture(scope="function")
def grid_data_d_1d(grid_data_d):
    return StructuredGrid1d(grid_data_d, "x")


@pytest.fixture(scope="function")
def grid_data_b_2d(grid_data_b):
    return StructuredGrid2d(grid_data_b, "x", "y")


@pytest.fixture(scope="function")
def grid_data_c_2d(grid_data_c):
    return StructuredGrid2d(grid_data_c, "x", "y")


@pytest.fixture(scope="function")
def grid_data_e_1d(grid_data_e):
    return StructuredGrid1d(grid_data_e, "x")


@pytest.fixture(scope="function")
def expected_results_centroid():
    return xr.DataArray(
        data=np.array(
            [
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                0,
                1,
                np.nan,
                np.nan,
                3,
                4,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            ]
        ).reshape((4, 4)),
        dims=["y", "x"],
        coords={
            "y": np.array([175, 125, 75, 25]),
            "x": np.array([25, 75, 125, 175]),
            "dx": 50.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def expected_results_overlap():
    # --------
    # target | source            | ntarget | sum(source)/ntarget
    # 0      | 0                 | 1       | 0.0
    # 1      | 0 1               | 2       | 0.5
    # 2      |   1 2             | 2       | 1.5
    # 3      |     2             | 1       | 2.0
    # 4      | 0     3           | 2       | 1.5
    # 5      | 0 1   3 4         | 4       | 2.0
    # 6      |   1 2   4 5       | 4       | 3.0
    # 7      |     2     5       | 2       | 3.5
    # 8      |       3     6     | 2       | 4.5
    # 9      |       3 4   6 7   | 4       | 5.0
    # 10     |         4 5   7 8 | 4       | 6.0
    # 11     |           5     8 | 2       | 6.5
    # 12     |             6     | 1       | 6.0
    # 13     |             6 7   | 2       | 6.5
    # 14     |               7 8 | 2       | 7.5
    # 15     |                 8 | 1       | 8.0
    # --------
    return xr.DataArray(
        data=np.array(
            [
                0.0,
                0.5,
                1.5,
                2.0,
                1.5,
                2.0,
                3.0,
                3.5,
                4.5,
                5.0,
                6.0,
                6.5,
                6.0,
                6.5,
                7.5,
                8.0,
            ]
        ).reshape((4, 4)),
        dims=["y", "x"],
        coords={
            "y": np.array([175, 125, 75, 25]),
            "x": np.array([25, 75, 125, 175]),
            "dx": 50.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def expected_results_linear():
    # --------
    # target | source            | weights                | sum(source x weight)
    # 5      | 0 1   3 4         | 0.25 0.25 0.25 0.25    | 2.0
    # 6      |   1 2   4 5       | 0.25 0.25 0.25 0.25    | 3.0
    # 9      |       3 4   6 7   | 0.25 0.25 0.25 0.25    | 5.0
    # 10     |         4 5   7 8 | 0.25 0.25 0.25 0.25    | 6.0
    # --------
    return xr.DataArray(
        data=np.array(
            [
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                2.0,
                3.0,
                np.nan,
                np.nan,
                5.0,
                6.0,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            ]
        ).reshape((4, 4)),
        dims=["y", "x"],
        coords={
            "y": np.array([175, 125, 75, 25]),
            "x": np.array([25, 75, 125, 175]),
            "dx": 50.0,
            "dy": -50.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_dask_source():
    data = np.arange(10000).reshape((100, 100))
    data = da.from_array(data, chunks=(10, (10, 30, 60)))
    return xr.DataArray(
        data=data,
        dims=["y", "x"],
        coords={
            "y": -np.arange(100),
            "x": np.arange(100),
            "dx": 1.0,
            "dy": -1.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_dask_source_layered():
    data = np.arange(30000).reshape((3, 100, 100))
    data = da.from_array(data, chunks=(3, 10, (10, 30, 60)))
    return xr.DataArray(
        data=data,
        dims=["layer", "y", "x"],
        coords={
            "layer": np.arange(3),
            "y": -np.arange(100),
            "x": np.arange(100),
            "dx": 1.0,
            "dy": -1.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_dask_target():
    data = np.zeros(100).reshape((10, 10))
    data = da.from_array(data, chunks=(10, (5, 5)))
    return xr.DataArray(
        data=data,
        dims=["y", "x"],
        coords={
            "y": -np.arange(10) * 10,
            "x": np.arange(10) * 10,
            "dx": 10.0,
            "dy": -10.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_dask_expected():
    data1 = np.tile(np.arange(0.0, 100.0, 10.0), reps=10).reshape((10, 10))
    data2 = np.repeat(np.arange(0.0, 10000.0, 1000.0), repeats=10).reshape((10, 10))
    data = data1 + data2
    return xr.DataArray(
        data=data,
        dims=["y", "x"],
        coords={
            "y": -np.arange(10) * 10,
            "x": np.arange(10) * 10,
            "dx": 10.0,
            "dy": -10.0,
        },
    )


@pytest.fixture(scope="function")
def grid_data_dask_expected_layered(grid_data_dask_expected):
    return grid_data_dask_expected.expand_dims(dim={"layer": np.arange(3)})



import os

import numpy as np
import pytest

from xugrid.core import sparse


def numba_enabled() -> bool:
    return os.environ.get("NUMBA_DISABLE_JIT") != "1"


@pytest.fixture(scope="function")
def coo_matrix():
    source_index = np.arange(10)
    target_index = np.repeat(np.arange(5), 2)
    weights = np.full(10, 0.5)
    return sparse.MatrixCOO.from_triplet(target_index, source_index, weights)


@pytest.fixture(scope="function")
def csr_matrix():
    source_index = np.arange(10)
    target_index = np.repeat(np.arange(5), 2)
    weights = np.full(10, 0.5)
    return sparse.MatrixCSR.from_triplet(target_index, source_index, weights)


def test_weight_matrix_coo(coo_matrix):
    assert isinstance(coo_matrix, sparse.MatrixCOO)
    assert np.allclose(coo_matrix.data, np.full(10, 0.5))
    assert np.array_equal(coo_matrix.row, [0, 0, 1, 1, 2, 2, 3, 3, 4, 4])
    assert np.array_equal(coo_matrix.col, np.arange(10))
    assert coo_matrix.nnz == 10


def test_weight_matrix_csr(csr_matrix):
    assert isinstance(csr_matrix, sparse.MatrixCSR)
    assert np.allclose(csr_matrix.data, np.full(10, 0.5))
    assert np.array_equal(csr_matrix.indices, np.arange(10))
    assert np.array_equal(csr_matrix.indptr, [0, 2, 4, 6, 8, 10])
    assert csr_matrix.n == 5
    assert csr_matrix.nnz == 10


@pytest.mark.skipif(
    numba_enabled(),
    reason="Numba cannot convert native range_state_int64 to Python object.",
)
def test_nzrange(csr_matrix):
    # These functions work fine if called inside of other numba functions when
    # numba is enabled.
    i = sparse.nzrange(csr_matrix, 0)
    assert np.array_equal(i, range(0, 2))
    i = sparse.nzrange(csr_matrix, 1)
    assert np.array_equal(i, range(2, 4))


@pytest.mark.skipif(
    numba_enabled(),
    reason="Function returns a slice object; python and no-python slices don't mix.",
)
def test_row_slice(csr_matrix):
    # These functions work fine if called inside of other numba functions when
    # numba is enabled.
    assert sparse.row_slice(csr_matrix, 0) == slice(0, 2, None)


@pytest.mark.skipif(
    numba_enabled(),
    reason="Function returns a zip object; python and no-python zips don't mix.",
)
def test_columns_and_values(csr_matrix):
    # These functions work fine if called inside of other numba functions when
    # numba is enabled.
    zipped = sparse.columns_and_values(csr_matrix, sparse.row_slice(csr_matrix, 0))
    result = list(zipped)
    assert result == [(0, 0.5), (1, 0.5)]


def test_coo_to_csr(coo_matrix):
    csr_matrix = coo_matrix.to_csr()
    assert isinstance(csr_matrix, sparse.MatrixCSR)
    assert np.allclose(csr_matrix.data, np.full(10, 0.5))
    assert np.array_equal(csr_matrix.indices, np.arange(10))
    assert np.array_equal(csr_matrix.indptr, [0, 2, 4, 6, 8, 10])
    assert csr_matrix.n == 5
    assert csr_matrix.nnz == 10


def test_csr_to_coo(csr_matrix):
    coo_matrix = csr_matrix.to_coo()
    assert isinstance(coo_matrix, sparse.MatrixCOO)
    assert np.allclose(coo_matrix.data, np.full(10, 0.5))
    assert np.array_equal(coo_matrix.row, [0, 0, 1, 1, 2, 2, 3, 3, 4, 4])
    assert np.array_equal(coo_matrix.col, np.arange(10))
    assert coo_matrix.nnz == 10


def test_shape():
    source_index = np.arange(10)
    target_index = np.repeat(np.arange(5), 2)
    weights = np.full(10, 0.5)
    matrix = sparse.MatrixCSR.from_triplet(target_index, source_index, weights, n=20)
    assert matrix.n == 20
    assert matrix.m == 10
    matrix = sparse.MatrixCSR.from_triplet(target_index, source_index, weights, m=20)
    assert matrix.n == 5
    assert matrix.m == 20



"""
Numba percentile methods allocate continuously on the heap.

This has significant overhead when calling the reduction method millions of
times -- as happens when regridding.

This is a simplified port of the percentile helpers:
# https://github.com/numba/numba/blob/0441bb17c7820efc2eba4fd141b68dac2afa4740/numba/np/arraymath.py#L1595
"""

import numba as nb
import numpy as np


@nb.njit(inline="always")
def nan_le(a, b) -> bool:
    # Nan-aware <
    if np.isnan(a):
        return False
    elif np.isnan(b):
        return True
    else:
        return a < b


@nb.njit
def _partition(A, low, high):
    mid = (low + high) >> 1
    # NOTE: the pattern of swaps below for the pivot choice and the
    # partitioning gives good results (i.e. regular O(n log n))
    # on sorted, reverse-sorted, and uniform arrays.  Subtle changes
    # risk breaking this property.

    # Use median of three {low, middle, high} as the pivot
    if nan_le(A[mid], A[low]):
        A[low], A[mid] = A[mid], A[low]
    if nan_le(A[high], A[mid]):
        A[high], A[mid] = A[mid], A[high]
    if nan_le(A[mid], A[low]):
        A[low], A[mid] = A[mid], A[low]
    pivot = A[mid]

    A[high], A[mid] = A[mid], A[high]
    i = low
    j = high - 1
    while True:
        while i < high and nan_le(A[i], pivot):
            i += 1
        while j >= low and nan_le(pivot, A[j]):
            j -= 1
        if i >= j:
            break
        A[i], A[j] = A[j], A[i]
        i += 1
        j -= 1
    # Put the pivot back in its final place (all items before `i`
    # are smaller than the pivot, all items at/after `i` are larger)
    A[i], A[high] = A[high], A[i]
    return i


@nb.njit
def _select(arry, k, low, high):
    """Select the k'th smallest element in array[low:high + 1]."""
    i = _partition(arry, low, high)
    while i != k:
        if i < k:
            low = i + 1
            i = _partition(arry, low, high)
        else:
            high = i - 1
            i = _partition(arry, low, high)
    return arry[k]


@nb.njit
def _select_two(arry, k, low, high):
    """
    Select the k'th and k+1'th smallest elements in array[low:high + 1].

    This is significantly faster than doing two independent selections
    for k and k+1.
    """
    while True:
        assert high > low  # by construction
        i = _partition(arry, low, high)
        if i < k:
            low = i + 1
        elif i > k + 1:
            high = i - 1
        elif i == k:
            _select(arry, k + 1, i + 1, high)
            break
        else:  # i == k + 1
            _select(arry, k, low, i - 1)
            break

    return arry[k], arry[k + 1]
